#!/usr/bin/env python

"""monchat client program."""

# David Fifield, 2007

# For getpass.getuser.
import getpass
import gtk
import socket
import threading

import message
from util import error, format_address

PORT = 2001

# The magic string which, when sent, causes us to disconnect.
MAGIC_QUIT_STRING = "Stop chatting. Learn JAVA or RUBY instead"

class MessageHandler(threading.Thread):
	"""A background thread that sends and receives messages. It writes received
	messages to a gtk.TextBuffer."""

	# How long to wait to read from the socket.
	SOCKET_TIMEOUT = 0.05

	def __init__(self, connection, main_window):
		threading.Thread.__init__(self)
		self.connection = connection
		self.main_window = main_window

		self.recv_buffer = ""
		self.send_queue = []
		self.quit = False

		self.send_queue_lock = threading.Lock()

		self.connection.settimeout(self.SOCKET_TIMEOUT)

	def run(self):
		try:
			while not self.quit:
				try:
					m = self.__read_message()
				except message.FormatError, e:
					error("Couldn't unformat message: %s." % str(e))
					self.quit = True
				except socket.timeout:
					pass
				else:
					if m is None:
						self.main_window.append_text(
							"*** Connection closed by remote host.\n")
						self.quit = True
					else:
						self.__handle_message(m)

				# Copy the send queue to prevent concurrent modification.
				self.send_queue_lock.acquire()
				tmp_queue = self.send_queue[:]
				self.send_queue = []
				self.send_queue_lock.release()

				# Send messages in the queue.
				while tmp_queue:
					self.__send_message(tmp_queue.pop(0))
		finally:
			self.connection.close()

	def stop(self):
		"""Stop the thread."""
		self.quit = True

	def queue_message(self, m):
		"""Add a message to the thread's send queue."""
		self.send_queue_lock.acquire()
		self.send_queue.append(m)
		self.send_queue_lock.release()

	def __send_message(self, m):
		"""Send a message."""
		try:
			self.connection.sendall(m.format())
		except socket.error, (errno, msg):
			self.main_window.append_text(msg + ".\n")

	def __handle_message(self, m):
		"""Handle a message by writing it to the gtk.TextBuffer."""
		if not m.header.has_key("Action") or m.header.get("Action") == "Message":
			# A broadcast message.
			self.main_window.append_text(m.header.get("ID") + ": " \
				+ m.body.strip() + "\n")
		else:
			# Server info.
			self.main_window.append_text("*** " + m.body.strip() + "\n")

	def __read_message(self):
		"""Return the next message extracted from the socket."""
		while True:
			head, sep, tail = self.recv_buffer.partition(message.Message.TERMINATOR)
			if sep:
				self.recv_buffer = tail
				return message.unformat(head + sep)
			data = self.connection.recv(1024)
			if data == "":
				return None
			self.recv_buffer += data

def get_default_id():
	"""Return a default ID. This should be the user's login name."""
	try:
		return getpass.getuser()
	except:
		return ""

class ConnectWindow(gtk.Window):
	"""A dialog window that prompts for a host and ID. When the "Connect"
	button is pressed, it opens a connection and displays a MainWindow."""
	TITLE = "Connect"
	DEFAULT_HOST = "localhost"

	LOGO_FILENAME = "logo.png"

	def __init__(self):
		gtk.Window.__init__(self)

		self.set_title(self.TITLE)
		self.set_border_width(10)
		self.set_resizable(False)
		self.connect("delete_event", gtk.main_quit)

		vbox = gtk.VBox(spacing = 6)

		logo = gtk.Image()
		logo.set_from_file(self.LOGO_FILENAME)
		logo.show()
		vbox.pack_start(logo)

		self.status_label = gtk.Label()
		self.status_label.show()
		vbox.pack_start(self.status_label, expand = False)

		table = gtk.Table(2, 2)
		table.set_row_spacings(6)
		table.set_col_spacings(6)

		label = gtk.Label("ID:")
		label.set_alignment(1.0, 0.5)
		label.show()
		table.attach(label, 0, 1, 0, 1)

		self.id_entry = gtk.Entry()
		self.id_entry.set_text(get_default_id())
		self.id_entry.set_activates_default(True)
		self.id_entry.show()
		table.attach(self.id_entry, 1, 2, 0, 1)

		label = gtk.Label("Host:")
		label.set_alignment(1.0, 0.5)
		label.show()
		table.attach(label, 0, 1, 1, 2)

		self.host_entry = gtk.Entry()
		self.host_entry.set_text(self.DEFAULT_HOST)
		self.host_entry.set_activates_default(True)
		self.host_entry.show()
		table.attach(self.host_entry, 1, 2, 1, 2)

		table.show()

		vbox.pack_start(table, expand = True, fill = False)

		hbox = gtk.HBox()

		button = gtk.Button("Connect")
		button.connect("clicked", self.connect_callback)
		button.show()

		button.set_alignment(1.0, 0.5)
		hbox.pack_end(button, expand = False)

		hbox.show()

		vbox.pack_start(hbox, expand = False)

		vbox.show()

		self.add(vbox)

		button.set_flags(gtk.CAN_DEFAULT)
		button.grab_default()

	def connect_callback(self, button):
		"""Try to connect."""
		host = self.host_entry.get_text()
		id = self.id_entry.get_text()

		self.status_label.set_text("Connecting to %s..." \
			% format_address((host, PORT)))
		try:
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.connect((host, PORT))
		except socket.error, (errno, msg):
			self.status_label.set_text(msg + ".")
		else:
			self.destroy()
			# Create and show the main window.
			main_window = MainWindow(s, id)
			main_window.show()

def loose_compare(s, t):
	"""Compare two strings loosely; get rid of punctuation and convert to
	lowercase before comparing."""
	s = "".join(c for c in s if c.isalnum()).lower()
	t = "".join(c for c in t if c.isalnum()).lower()

	return s == t

class MainWindow(gtk.Window):
	"""The main application window that displays the chat."""
	TITLE = "Chat window"
	WIDTH = 600
	HEIGHT = 600

	# The width, in characters, of the ID entry field.
	ID_ENTRY_CHARS = 12

	def __init__(self, connection, id = ""):
		gtk.Window.__init__(self)

		self.set_title(self.TITLE)
		self.set_default_size(self.WIDTH, self.HEIGHT);
		self.set_border_width(10)
		self.connect("delete_event", self.quit);

		vbox = gtk.VBox(spacing = 4)

		hbox = gtk.HBox(spacing = 6)
		label = gtk.Label("Current ID")
		label.show()
		hbox.pack_start(label, expand = False)
		self.id_entry = gtk.Entry()
		self.id_entry.set_width_chars(max(len(id), self.ID_ENTRY_CHARS))
		self.id_entry.set_text(id)
		self.id_entry.connect("activate", self.id_entry_activate)
		self.id_entry.show()
		hbox.pack_start(self.id_entry, expand = False)
		label = gtk.Label()
		label.set_markup(
			"<small>(Changing this will change your chat ID.)</small>")
		label.show()
		hbox.pack_start(label, expand = False)
		hbox.show()

		vbox.pack_start(hbox, expand = False)

		scrolled_window = gtk.ScrolledWindow()
		scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		scrolled_window.show()
		vbox.pack_start(scrolled_window, expand = True, fill = True)

		self.textview = gtk.TextView()
		self.textview.set_editable(False)
		self.textview.set_cursor_visible(False)
		self.textview.set_wrap_mode(gtk.WRAP_WORD)
		self.textview.show()
		scrolled_window.add(self.textview)

		self.textbuffer = self.textview.get_buffer()

		input_entry = gtk.Entry()
		input_entry.connect("activate", self.input_entry_activate)
		input_entry.show()
		vbox.pack_start(input_entry, expand = False)

		vbox.show()

		self.add(vbox)

		input_entry.grab_focus()

		self.handler = MessageHandler(connection, self)
		self.handler.start()

		# Send a message announcing we've entered the chat.
		m = message.Message()
		m.header["ID"] = id
		m.header["Action"] = "Enter"
		self.handler.queue_message(m)

	def input_entry_activate(self, entry):
		"""Send the message that's in the input text entry."""
		text = entry.get_text().strip()
		id = self.id_entry.get_text()

		# Check if the user typed the magic quit string, or something close to
		# it.
		if loose_compare(text, MAGIC_QUIT_STRING):
			m = message.Message()
			m.header["ID"] = id
			m.header["Action"] = "Leave"
			m.body = text
		else:
			m = message.Message()
			m.header["ID"] = id
			m.body = text

		self.handler.queue_message(m)

		entry.set_text("")

	def id_entry_activate(self, entry):
		"""Set the ID in the ID text entry."""
		m = message.Message()
		m.header["ID"] = self.id_entry.get_text()
		m.header["Action"] = "Enter"
		self.handler.queue_message(m)

	def append_text(self, text):
		"""Append text to the chat screen."""
		gtk.gdk.threads_enter()
		self.textbuffer.insert(self.textbuffer.get_end_iter(), text)
		self.textview.scroll_mark_onscreen(self.textbuffer.get_insert())
		gtk.gdk.threads_leave()

	def quit(self, source, event):
		"""Stop the message handler and quit the program."""
		self.handler.stop()
		self.handler.join()
		gtk.main_quit()

		return False

def main():
	connect_window = ConnectWindow()
	connect_window.show()

	gtk.gdk.threads_init()
	gtk.gdk.threads_enter()
	gtk.main()
	gtk.gdk.threads_leave()

if __name__ == "__main__":
	main()
