"""Client handler. This module contains a class ClientHandler, an instance of
which is used by the server to dispatch and receive message from a client. This
module also keeps its own list of all connected clients."""

# David Fifield, 2007

import socket
import threading

import message
from util import debug_print, error, format_address

client_list = []

# A lock on setting client IDs, to keep two clients from setting the same ID at
# the same time.
id_lock = threading.Lock()

def broadcast_message(m):
	"""Send a message to all connected clients that have an ID."""
	id_lock.acquire()
	clients = [client for client in client_list if client.id is not None]
	id_lock.release()

	for client in clients:
		client.queue_message(m)

class ClientHandler(threading.Thread):
	"""A client handling thread.

	One of these threads is created for client connection. The thread
	dispatches message between the client and server.
	"""

	# How long to wait for incoming data on the socket each time through the
	# loop.
	SOCKET_TIMEOUT = 0.05

	# How much incoming data to read at a time.
	RECV_BUFFER_SIZE = 1024

	def __init__(self, connection, address):
		"""Initialize the client handler.

		connection must be an open socket. address is used in messages to
		identify this handler.
		"""
		threading.Thread.__init__(self)
		self.connection = connection
		self.address = address
		self.id = None

		self.recv_buffer = ""
		self.send_queue = []
		self.quit = False

		self.send_queue_lock = threading.Lock()

		self.connection.settimeout(self.SOCKET_TIMEOUT)

		client_list.append(self)

		debug_print("Client handler initialized for %s." \
			% format_address(self.address))

	def run(self):
		"""Run the main loop of the thread.

		This periodically checks for new data on the socket and interprets it as
		messages. It then sends messages that are in its send queue.
		"""
		debug_print("Client handler for %s started." \
			% format_address(self.address))
		try:
			m = message.Message()
			m.header["Action"] = "Info"
			m.body = "Hi %s. Welcome to monchat!" % self.address[0]
			self.queue_message(m)

			while not self.quit:
				try:
					m = self.__read_message()
				except message.FormatError, e:
					error("Couldn't unformat message from %s: %s" \
						% (format_address(self.address), str(e)))
					self.quit = True
				except socket.timeout:
					pass
				else:
					if m is None:
						self.quit = True
					else:
						self.__handle_message(m)

				# Copy the send queue to prevent concurrent modification.
				self.send_queue_lock.acquire()
				tmp_queue = self.send_queue[:]
				self.send_queue = []
				self.send_queue_lock.release()

				# Send all the messages in the send queue.
				while tmp_queue:
					self.__send_message(tmp_queue.pop(0))
		except Exception, e:
			debug_print("Client handler for %s caught exception: %s" \
				% (format_address(self.address), str(e)))
		finally:
			client_list.remove(self)
			# Broadcast a message when someone leaves the chat. Don't bother
			# notifying other clients when a client without an ID leaves.
			if self.id is not None:
				broadcast = message.Message()
				broadcast.header["Action"] = "Leave"
				broadcast.header["ID"] = self.id
				broadcast.body = "\"%s\" has left the chat.\n" % self.id
				broadcast_message(broadcast)
			debug_print("Client handler for %s closing socket." \
				% format_address(self.address))
			self.connection.close()

	def queue_message(self, m):
		"""Add a message to the handler's send queue."""
		self.send_queue.append(m)

	def __send_message(self, m):
		"""Send a message on the handler's socket."""
		formatted = m.format()
		debug_print("Sending to %s:" % format_address(self.address))
		debug_print("\n".join("  >> " + line for line in formatted.splitlines()))
		self.connection.sendall(formatted)

	def __handle_message(self, m):
		"""Dispatch a message received from the client."""
		debug_print("Received from %s:" % format_address(self.address))
		debug_print("\n".join("<<   " + line for line in m.format().splitlines()))

		# Set or change the ID.
		if m.header.has_key("ID"):
			if self.id is None:
				if self.__try_change_id(m.header["ID"]):
					debug_print("Client at %s is known as \"%s\"." \
						% (format_address(self.address), self.id))
			elif m.header["ID"] != self.id:
				if self.__try_change_id(m.header["ID"]):
					debug_print("Client at %s is now known as \"%s\"." \
						% (format_address(self.address), self.id))

		# A client normally must set its ID before sending any messages. The
		# only exception is a message with the "Leave" action.
		if m.header.get("Action") == "Leave":
			self.quit = True
			reply = message.Message()
			reply.header["Action"] = "Info"
			if self.id is not None:
				name = self.id
			else:
				name = self.address[0]
			reply.body = "So long, %s. It was nice knowing you.\n" % name
			self.queue_message(reply)
			return

		if self.id is None:
			reply = message.Message()
			reply.header["Action"] = "Info"
			reply.body = "You must set your ID before sending any messages.\n"
			self.queue_message(reply)
			return

		if not m.header.has_key("Action") or m.header["Action"] == "Message":
			# Ensure that the message has an "ID" header field.
			m.header["ID"] = self.id
			broadcast_message(m)
		elif m.header.get("Action") == "Enter":
			# Don't do anything. Allow performing this action multiple times.
			# This allows a client to announce it has entered the chat without
			# sending a message.
			pass
		else:
			# An unknown action. Kick the client off with an error message.
			self.quit = True
			reply = message.Message()
			reply.header["Action"] = "Info"
			reply.body = "Unknown action: \"%s\".\n" % m.header.get("Action")
			self.queue_message(reply)

	def __try_change_id(self, newid):
		"""Attempt to change this client handler's ID.

		This first checks to see that no other handler is using the ID. If one
		is, a message is sent to the client that requested the ID change. If
		the name can be changed, a broadcast message is sent informing all
		clients of the change.
		"""
		id_lock.acquire()
		in_use = newid in [client.id for client in client_list]
		id_lock.release()

		if in_use:
			reply = message.Message()
			reply.header["Action"] = "Info"
			reply.body = "The ID \"%s\" is already in use." % newid
			self.queue_message(reply)
			return False
		else:
			oldid = self.id
			self.id = newid
			broadcast = message.Message()
			if oldid is None:
				broadcast.header["Action"] = "Enter"
				broadcast.header["ID"] = self.id
				broadcast.body = "\"%s\" has entered the chat.\n" % self.id
			else:
				broadcast.header["Action"] = "Change-ID"
				broadcast.header["Old-ID"] = oldid
				broadcast.header["ID"] = self.id
				broadcast.body = "\"%s\" is now known as \"%s\".\n" \
					% (oldid, self.id)
			broadcast_message(broadcast)
			return True

	def __read_message(self):
		"""Return the next message extracted from the socket."""
		while True:
			head, sep, tail = self.recv_buffer.partition(message.Message.TERMINATOR)
			if sep:
				self.recv_buffer = tail
				return message.unformat(head + sep)
			data = self.connection.recv(1024)
			if data == "":
				return None
			self.recv_buffer += data
