#!/usr/bin/env python

import select
import socket
import unittest
import time

import clienthandler
import message
import util

util.DEBUG = False

class FakeSocket:
	REMOTE_SLEEP_TIME = 0.05

	def __init__(self):
		self.recv_buffer = ""
		self.send_buffer = ""
		self.timeout = 0.0
		self.closed = False
		self.remote_closed = False

		self.message_buffer = ""

	def settimeout(self, timeout):
		self.timeout = timeout

	def recv(self, n):
		if not self.recv_buffer and not self.remote_closed:
			time.sleep(self.timeout)
			raise socket.timeout
		if self.recv_buffer:
			tmp = self.recv_buffer[:n]
			self.recv_buffer = self.recv_buffer[n:]
			return tmp
		else:
			return ""

	def sendall(self, data):
		self.send_buffer += data

	def close(self):
		self.closed = True

	def remote_recv(self):
		"""Simulate receiving data to the remote host."""
		while not self.send_buffer and not self.closed:
			time.sleep(self.REMOTE_SLEEP_TIME)
		tmp = self.send_buffer
		self.send_buffer = ""
		return tmp

	def remote_send(self, data):
		"""Simulate sending data from the remote host."""
		self.recv_buffer += data

	def remote_close(self):
		"""Simulate closing the socket from the remote host."""
		self.remote_closed = True

	def remote_send_message(self, m):
		self.remote_send(m.format())

	def remote_recv_message(self):
		while True:
			head, sep, tail = self.message_buffer.partition(message.Message.TERMINATOR)
			if sep:
				self.message_buffer = tail
				return message.unformat(head + sep)
			data = self.remote_recv()
			if data == "":
				return None
			self.message_buffer += data

	def __iter__(self):
		return self

	def next(self):
		m = self.remote_recv_message()
		if m is None:
			raise StopIteration
		return m

def make_handler(host = "localhost"):
	s = FakeSocket()
	handler = clienthandler.ClientHandler(s, (host, 0))
	handler.setDaemon(True)
	handler.start()

	return handler

class ClientHandlerTest(unittest.TestCase):
	ID1 = "Nobody"
	ID2 = "Dummy"
	BODY = "Hello."

	def test_set_id(self):
		"""Test that sending a message with the "ID" header sets the client's
		ID and enters the chat."""
		h = make_handler()
		m = message.Message({"ID": self.ID1})
		h.connection.remote_send_message(m)
		h.connection.remote_close()
		h.join()
		for m in h.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID1:
				break
		else:
			fail("Did not receive an \"Enter\" notification.")
		self.assertEqual(self.ID1, h.id)

	def test_change_id(self):
		"""Test that changing an ID works."""
		h = make_handler()
		m = message.Message({"ID": self.ID1})
		h.connection.remote_send_message(m)
		m = message.Message({"ID": self.ID2})
		h.connection.remote_send_message(m)
		h.connection.remote_close()
		h.join()
		for m in h.connection:
			if m.header.get("Action") == "Change-ID":
				break
		else:
			self.fail("Did not receive a message with a \"Change-ID\" action.")
		self.assertEqual(self.ID2, h.id)

	def test_no_dup_id(self):
		"""Test that duplicate IDs are rejected."""
		h1 = make_handler("client1")
		m = message.Message({"ID": self.ID1, "Action": "Enter"})
		h1.connection.remote_send(m.format())

		# Wait until the ID is set.
		for m in h1.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID1:
				break

		h2 = make_handler("client2")
		m = message.Message({"ID": self.ID1, "Action": "Enter"})
		h2.connection.remote_send_message(m)
		h2.connection.remote_close();
		h2.join()

		h1.connection.remote_close();
		h1.join()

		self.assertFalse(self.ID1 == h1.id and self.ID1 == h2.id)

	def test_leave_no_id(self):
		"""Test that the server allows sending a message with the "Leave" action
		without first setting an ID."""
		h = make_handler()
		m = message.Message({"Action": "Leave"})
		h.connection.remote_send_message(m)
		# If the test doesn't hang, it passes.
		h.join()

	def test_broadcast(self):
		"""Test that messages are broadcast."""
		h1 = make_handler("client1")
		m = message.Message({"ID": self.ID1, "Action": "Enter"})
		h1.connection.remote_send_message(m)
		for m in h1.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID1:
				break

		h2 = make_handler("client2")
		m = message.Message({"ID": self.ID2, "Action": "Enter"})
		h2.connection.remote_send_message(m)
		for m in h1.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID2:
				break

		m = message.Message({"ID": self.ID1}, self.BODY)
		h1.connection.remote_send_message(m)
		h1.connection.remote_close()
		h1.join()
		for m in h1.connection:
			if m.header.get("ID") == self.ID1 and m.body == self.BODY:
				break
		else:
			self.fail("Did not receive a broadcast message from self.")

		h2.connection.remote_close()
		h2.join()
		for m in h2.connection:
			if m.header.get("ID") == self.ID1 and m.body == self.BODY:
				break
		else:
			self.fail("Did not receive a broadcast message from another user.")

	def test_no_anonymous_send(self):
		"""Test that messages are not broadcast from clients that have not set
		an ID."""
		h1 = make_handler("client1")
		m = message.Message({"ID": self.ID1, "Action": "Enter"})
		h1.connection.remote_send_message(m)
		for m in h1.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID1:
				break

		h2 = make_handler("client2")
		m = message.Message({}, self.BODY)
		h2.connection.remote_send_message(m)
		h2.connection.remote_close()
		h2.join()

		h1.connection.remote_close()
		h1.join()
		for m in h1.connection:
			if m.body == self.BODY:
				self.fail("Received a broadcast message from an anonymous client.")

	def test_no_anonymous_receive(self):
		"""Test that messages are not broadcast to clients that have not set an
		ID."""
		h1 = make_handler("client1")
		m = message.Message({"ID": self.ID1, "Action": "Enter"})
		h1.connection.remote_send_message(m)
		for m in h1.connection:
			if m.header.get("Action") == "Enter" and m.header.get("ID") == self.ID1:
				break

		h2 = make_handler("client2")

		m = message.Message({}, self.BODY)
		h1.connection.remote_send_message(m)
		h1.connection.remote_close()
		h1.join()

		h2.connection.remote_close()
		h2.join()
		for m in h2.connection:
			if m.header.get("ID") == self.ID1 and m.body == self.BODY:
				self.fail("Anonymous client received a broadcast message.")

if __name__ == "__main__":
	unittest.main()
