#!/usr/bin/env python

import unittest

import message

class MessageTest(unittest.TestCase):
	SAMPLE_HEADER = {"ID": "user", "Action": "Send", "Bad": "a\nb"}
	SAMPLE_BODY = u"""
A sample body containing a blank line at the beginning,
strange characters: %$#"!*()\\,
Non-ASCII characters (which are encoded to byte sequences): \u6574 \u0153

an embedded blank line,
and something that looks like an encoded newline: %0A.
(Also the last line ends with a newline.)
""".encode("utf_8")

	def setUp(self):
		self.m = message.Message(self.SAMPLE_HEADER, self.SAMPLE_BODY)

	def test_equality(self):
		"""Test that two messages created with the same data compare equal."""
		self.assertEqual(message.Message(), message.Message())
		other = message.Message(self.SAMPLE_HEADER, self.SAMPLE_BODY)
		self.assertEqual(self.m, other)

	def test_terminator(self):
		"""Test that a formatted message contains exactly one "\n.\n"
		substring, at the end."""
		formatted = self.m.format()
		self.assertEqual(formatted.count(message.Message.TERMINATOR), 1)
		self.assertTrue(formatted.endswith(message.Message.TERMINATOR))

	def test_header_body_separator(self):
		"""Test that a formatted message contains at least one "\n\n"
		substring."""
		formatted = self.m.format()
		self.assertNotEqual(formatted.find(message.Message.SEPARATOR), -1)

	def test_unformat(self):
		"""Test that attempting to unformat an improperly formatted message
		raises an exception."""
		# No separator, no terminator.
		self.assertRaises(message.FormatError, message.unformat, "")
		# No separator
		self.assertRaises(message.FormatError, message.unformat, "\n.\n")
		# No terminator.
		self.assertRaises(message.FormatError, message.unformat, "\n\n")
		# Invalid escaped character.
		self.assertRaises(message.FormatError, message.unformat, "%2@\n\n")
		# Escaped character spans end of string.
		self.assertRaises(message.FormatError, message.unformat, "\n\n%2")

	def test_reverse(self):
		"""Test that formatting then unformatting a message produces the
		original message."""
		# Do an empty message.
		empty = message.Message()
		self.assertEqual(message.unformat(message.Message().format()),
			message.Message())
		# Do a non-empty message.
		self.assertEqual(message.unformat(self.m.format()), self.m)

	def test_unicode(self):
		"""Test that unencoded Unicode strings with non-ASCII characters are
		rejected."""
		UNICODE_STRING = u"abc\u6574"
		try:
			message.Message({UNICODE_STRING: UNICODE_STRING}, UNICODE_STRING)
			self.fail()
		except:
			pass
		try:
			m.body = UNICODE_STRING
			self.fail()
		except:
			pass
		try:
			m.header["a"] = UNICODE_STRING
			self.fail()
		except:
			pass
		try:
			m.header[UNICODE_STRING] = "a"
			self.fail()
		except:
			pass

	def test_header_field_set(self):
		"""Test setting and getting the value of a header field."""
		self.m.header["Test"] = " Test data"
		self.assertEqual(self.m.header["Test"], " Test data")

	def test_format_blank(self):
		"""Test that formatting a blank message does not raise an exception."""
		m = message.Message({}, "")
		m.format()
		# Test formatting a blank header field value.
		m.header["ID"] = ""
		m.format()

if __name__ == "__main__":
	unittest.main()
